#!/usr/bin/perl 

#
# Install script for vmail-sql
#
# (c) 2001 Paul Warren
#
# $Id: install.pl,v 1.9 2001/10/01 21:01:02 pdw Exp $
#

use Term::Complete;
use Data::Dumper; # for debug only
use FileHandle;
use Getopt::Std;

#
# TODO: sort out sample vs debug
#
@options = (
    {
        name => 'sqluser',
        description => 'SQL username',
        sample => '""',
    },
    {
        name => 'sqlpasswd',
        description => 'SQL password',
        sample => '""',
    },
    {
        name => 'sqldb',
        description => 'SQL database',
        sample => '""',
    },
    {
        name => 'install_cgi_dir',
        description => 'Specify your cgi-bin directory.  Leave as undef if you do not wish to install the web front-end.',
        sample => 'undef',
    },
    {
        name => 'install_web_docs_dir',
        description => 'Web documents directory.  Absolute path to directory to store static web content (stylesheets, images).  Leave as undef if you do not iwsh to install the web front-end.',
        sample => 'undef',
        default => sub {
            my $dir = $options{install_cgi_dir}->{value};
            if ($dir) {
                $dir =~ s|[^/]+$|html|;
            } else {
                $dir = '';
            }
            return $dir;
        }
    },
    {
        name => 'install_lib_dir',
        description => <<EOF,
Directory for shared perl modules.  This directory contains perl modules used 
by both web and command line scripts.  These contain configuration options, 
including your database username and password.
EOF
        sample => 'undef',
        default => sub {
            my $dir = $options{install_cgi_dir}->{value};
            if ($dir) {
                $dir =~ s|[^/]+$|lib|;
            } else {
                $dir = '';
            }
            return $dir;
        }
    },
    {
        name => 'lib_dir_mode',
        description => 'Mode for lib dir.  This should probably be 0700 if you use Apache\'s suExec or 0755 otherwise.',
        sample => '"0700"',
        default => sub { return '0700' }
    },
    {
        name => 'install_bin_dir',
        description => 'Directory for command-line administrative tools.',
        version => 0.4,
        sample => 'undef',
        default => sub {
            my $dir = $options{install_lib_dir}->{value};
            if ($dir) {
                $dir =~ s|[^/]+$|bin|;
            } else {
                $dir = '';
            }
            return $dir;
        }
    },
    {
        name => 'mail_spool_path',
        description => 'Path under which virtual mail spools will be created.',
        sample => '"/var/spool/mail/SERVERS/"',
    },
    {
        name => 'default_hash_method',
        description => 'Sets the hashing method used to store passwords in the database.  Currently includes md5, crypt, crypt-md5 and plaintext.  See README for more details.',
        sample => '"md5"',	  
	default => sub { return "md5" },
    },
    {
        name => 'allow_plaintext_password',
        description => 'Allows the user to choose to have their password stored in plaintext in the database.  This is necessary if the user wished to use APOP authentication, but otherwise storing plaintext passwords is undesirable.',
        sample => '0',	  
	default => sub { return 0 },
    },
    {
        name => 'perl_bin',
        description => 'Perl binary location',
        sample => '"/usr/bin/perl"',
        default => sub {
            my $perl_bin = `which perl`;
            chomp($perl_bin);
            return $perl_bin;
        }
    },
    {
        name => 'css',
        description => 'Style sheet file',
        sample => '"/std.css"',
        default => sub {
            return "/std.css";
        }
    },
    {
        name => 'version',
        description => 'Version',
        sample => '0.0'
    }
); 


#
# @changes: a list of changes required for given a version.
#
@changes = (
    {
        version => '0.4',
        name => 'Database upgrade required',
        description => 'The database format for storing password hashes has changed.  Please review and apply db/upgrade-0.4.sql before using this version of vmail.'
    },
    {
        version => '0.5',
        name => 'Database upgrade required',
        description => 'The database tables for popboxes and forwarders are altered slightly for this release. Please review and apply db/upgrade-0.5.sql before using this version of vmail.'
    },
    {
        version => '0.5',
        name => 'Exim config update required',
        description => 'The Exim configuration has been altered for this release to support aliasing of domains.  Please see the config in the README file.'
    }
);

#
# hash_options: creates a hash of options name.  We store them in an array
# to preserve order, and hash them for ease of access.
#
sub hash_options {
    my %options;
    foreach my $option (@options) {
        $options{$option->{name}} = $option;
    }
    return %options;
}

#
# word_wrap(text, start):
#
sub word_wrap {
    my ($text, $start) = @_;
    my $out;
    $text =~ tr/\n/ /;
    my $ll = 80 - length($start);
    while ($text) {
        if ($text =~ s/^(.{1,$ll})(?:\s+|$)//) {
            $out .= "$start$1\n"; 
        } elsif ($text =~ s/^\s*(\S+)//) {
            $out .= "$start$1\n";
        } else {
            last;
        }
    }
    return $out;
}

#
# ask_yn_question:
#
sub ask_yn_question {   
    my $question = shift;
    my $answer;
    do {
        $answer = Complete($question." (y/n) ");
    } while ($answer !~ /^[yn]$/i);
    return($answer =~ /y/i);
}

#
# ask_config_option
#
sub ask_config_option {
    my $option = shift;
    my ($default, $question);
    print "\n".$option->{description}."\n";
    $question = $option->{name}.": ";
    if ($option->{default}) {
        $default = &{$option->{default}};
        $question .= "[$default] "; 
    }
    $answer = Complete($question);

    if ($answer eq '') {
        $option->{value} = $default;
    } else {
        $option->{value} = $answer;
    }
}


#
# escape_scalar:
#
sub escape_scalar {
    $_ = shift;
    s/\\/\\\\/g;
    s/'/\'/g;
    return "'$_'";
}

#
# read_old_options:
#
sub read_old_options {
    my $value;
    foreach my $option (@options) {
        # This code doesn't distinguish between the undef "value" and
        # config options which not defined.
        $value = eval('$DomainAdminConfig::' . $option->{name} );
        
        if (defined($value)) {
            $option->{value} = $value;
        }
    }
}

#
# parse_config_file:
#
sub parse_config_file {
    my $config_file = shift;
    my $fin = new FileHandle;
    if (!$fin->open("<$config_file")) {
        print STDERR "Could not open $config_file: $!\n";
        exit(1);
    }

    my $config_data;
    while (<$fin>) {
        $config_data .= $_;
    }
    $fin->close() or die("Error closing file: $!\n");

    
    if (!defined(eval($config_data))) {
        print STDERR "Error running DomainAdminConfig.pm: $@\n";
        exit(1);
    }

    read_old_options();
}

#
# rewrite_perl_file: Rewrite the #! line of a perl file to include the 
# correct -I include line.
#
sub rewrite_perl_file {
    my $file = shift;
    my $fin = new FileHandle("<$file.in") 
        or die("Could not open $file: $!");
    my $fout = new FileHandle(">$file") 
        or die("Could not open $file: $!");

    my $line = <$fin>;
    if ($line =~ s/^#!\S+\s+//) {
        chomp($line);
        $line =~ s/(?:^|\s)-I(?:\S+|\s+\S+)//g;
        $line = '#!'.$options{perl_bin}->{value}." ".$line;
        $line .= " -I".$options{install_lib_dir}->{value}."\n";
    }

    $fout->print($line);

    while (<$fin>) {
        $fout->print($_); 
    }
    $fin->close();
    $fout->close();
}

#
# prepare_files
#
sub prepare_files {
    foreach $dir ("scripts","web/cgi-bin") {
        opendir(DIR, $dir) 
            or die("Can't open directory $dir: $!");
        foreach my $file (grep { /\.in$/ } readdir(DIR)) {
            if (-f "$dir/$file") {
                $file =~ s/\.in$//;
                print "Rewriting $dir/$file\n";
                rewrite_perl_file("$dir/$file"); 
            }
        }
        closedir(DIR);
    }
}

#
# copy_dir:
#
sub copy_dir {
    my ($from, $to, $mode) = @_;

    if (!$mode) {
        $mode = '0755';
    }

    my $fin = new FileHandle("<$from/FILES") 
        or die("Couldn't open $from/FILES");

    my @from_files = map { chomp; "$from/$_" } <$fin>;
    $fin->close();

    if (!@from_files) {
        print STDERR "No files found in $from/FILES\n";
        return undef;
    }

    if (!$to) {
        die("No destination directory specified when copying files from: $from");
    }

    if (! -d $to) {
        system("mkdir -m $mode -p $to ") == 0
            or die("Could not create directory: $to");
    }

    system("install -m $mode ".join(" ",@from_files)." $to") == 0
        or die("Could not copy from: $from to: $to");
}

#
# copy_files:
#
sub copy_files {
    copy_dir("lib",$options{install_lib_dir}->{value},$options{lib_dir_mode}->{value});
    copy_dir("scripts",$options{install_bin_dir}->{value});
    copy_dir("web/cgi-bin",$options{install_cgi_dir}->{value});
    copy_dir("web/docs",$options{install_web_docs_dir}->{value});
}

#
# write_config_module:
#
sub write_config_module {
    my $sample = shift;
    my $filename;
    if (!$sample) {
        $filename = "lib/DomainAdminConfig.pm";
    } else {
        $filename = "Config.pm";
    }

    my $fout = new FileHandle(">$filename")
        or die("Could not write $filename: $!");
    $fout->print('#!'.$options{perl_bin}->{value}."\n\n");
    $fout->print("package DomainAdminConfig;\n\n");
    foreach my $option (@options) {
        $fout->print("#\n".word_wrap($option->{description},"# ")."#\n");
        $fout->print('$'.$option->{name}." = ");
        if (!$sample) {
            $fout->print(escape_scalar($option->{value}))
        } else {
            $fout->print($option->{sample});
        }
        $fout->print(";\n\n"); 
    }
    $fout->close();
}

sub list_changes {
    print "\n";
    foreach my $change (sort { $a->{version} cmp $b->{version} } @changes) {
        if ($change->{version} le $old_version) { next };
        print "($change->{version}) $change->{name}\n";
        print word_wrap($change->{description});
        print "\n";
    }
}

########################################################################
#
# main:
#

#
# Create a hash of options for ease of access
#
%options = hash_options();

getopts('p');

if ($opt_p) {
    #
    # p = pre-dist.  Produces a Config.pm for distribution
    # 
    $options{perl_bin}->{value} = '/usr/bin/perl'; 
    write_config_module(1);
    exit;
}

if (@ARGV) {
    #
    # If a file is specified, we perform an upgrade based on that file
    #
    $upgrade = 1;

    my $config_file = shift(@ARGV);

    if (!-e $config_file) {
        print "Could not find $config_file\n";
        exit(1);
    }

    # Be smart about whether we're given a directory or file
    if (-d $config_file) {
        $config_file .= "/DomainAdminConfig.pm";
    }

    print "Upgrading based on $config_file...\n";
    parse_config_file($config_file);

    # Set previous version
    if(defined($DomainAdminConfig::version)) {
        $old_version = $DomainAdminConfig::version ;
        print "Found version: $old_version\n";
    }
    else { 
        $old_version = 0;
    }
} else {
    #
    # Install
    #
    
    $upgrade = 0;
    parse_config_file("Config.pm");

    if (!defined($options{install_lib_dir}->{value})) {
        print "Please edit Config.pm before running $0.\n";
        print "You did read the instructions, didn't you?\n";
        exit(1);
    }
    $old_version = 0;
}

#
# Read in version of this distribution
#
my $fin = new FileHandle "<VERSION" or die("Could not read VERSION: $!\n");
$new_version = <$fin>;
$fin->close() or die ("Error closing file: $!\n");

$options{version}->{value} = $new_version;

# Ask relevant upgrade/install questions
foreach my $option (@options) {
    if (!defined($option->{value})) {
        ask_config_option($option);
    }
}

prepare_files();
write_config_module();
copy_files();
list_changes();
